# ruff: noqa: N802 N803
import io
import typing as ta

from . import runtime as antlr4


##


def pformat(
        node: ta.Any,
        *,
        buf: ta.IO | None = None,
        indent: str = '',
        child_indent: str = '  ',
) -> ta.IO:
    if buf is None:
        buf = io.StringIO()
    buf.write(indent)
    buf.write(node.__class__.__name__)
    if hasattr(node, 'start') and hasattr(node, 'stop'):
        buf.write(f' ({node.start} -> {node.stop})')
    buf.write('\n')
    for child in getattr(node, 'children', []) or []:
        pformat(child, buf=buf, indent=indent + child_indent, child_indent=child_indent)
    return buf


def yield_contexts(
        root: antlr4.ParserRuleContext,
) -> ta.Iterator[antlr4.ParserRuleContext]:
    q = [root]
    while q:
        c = q.pop()
        yield c
        if not isinstance(c, antlr4.TerminalNode) and c.children:
            q.extend(c.children)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import collections.abc
import dataclasses as dc
import typing as ta

from ...lite.check import check
from ...lite.types import BUILTIN_SCALAR_ITERABLE_TYPES


T = ta.TypeVar('T')
StrT = ta.TypeVar('StrT', bound=str)
MappingT = ta.TypeVar('MappingT', bound=ta.Mapping)
IterableT = ta.TypeVar('IterableT', bound=ta.Iterable)

ConfigRewriterItem = ta.Union[int, str, None]  # ta.TypeAlias
ConfigRewriterPath = ta.Tuple[ConfigRewriterItem, ...]  # ta.TypeAlias


##


class RawConfigMetadata:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError


class ConfigRewriter(abc.ABC):  # noqa
    @dc.dataclass(frozen=True)
    class Context(ta.Generic[T]):
        obj: T

        parent: ta.Optional['ConfigRewriter.Context'] = None
        item: ConfigRewriterItem = None

        raw: bool = False

        def __post_init__(self) -> None:
            if self.parent is None:
                check.none(self.item)

        @property
        def path(self) -> ConfigRewriterPath:
            cur: ConfigRewriter.Context = self
            lst: ta.List[ConfigRewriterItem] = []
            while True:
                if cur.parent is None:
                    break
                lst.append(cur.item)
                cur = cur.parent
            return tuple(reversed(lst))

        def make_child(
                self,
                obj: ta.Any,
                item: ConfigRewriterItem,
                **kwargs: ta.Any,
        ) -> 'ConfigRewriter.Context':
            return dc.replace(
                self,
                obj=obj,
                parent=self,
                item=item,
                **kwargs,
            )

    def rewrite_none(self, ctx: Context[None]) -> ta.Any:
        return ctx.obj

    def rewrite_dataclass(self, ctx: Context[T]) -> T:
        kw = {}
        for f in dc.fields(ctx.obj):  # type: ignore
            fv = getattr(ctx.obj, f.name)
            nfv: ta.Any = self.rewrite(ctx.make_child(
                fv,
                f.name,
                raw=bool(not f.metadata.get(RawConfigMetadata)),
            ))
            if fv is not nfv:
                kw[f.name] = nfv
        if not kw:
            return ctx.obj
        return dc.replace(ctx.obj, **kw)  # type: ignore

    def rewrite_str(self, ctx: Context[StrT]) -> StrT:
        return ctx.obj

    def rewrite_builtin_scalar_iterable(self, ctx: Context[IterableT]) -> IterableT:
        return ctx.obj

    def rewrite_mapping(self, ctx: Context[MappingT]) -> MappingT:
        nm = []
        b = False
        for mk, mv in ctx.obj.items():
            nk: ta.Any = self.rewrite(ctx.make_child(mk, None))
            nv: ta.Any = self.rewrite(ctx.make_child(mv, nk))
            nm.append((nk, nv))
            b |= nk is not mk or nv is not mv
        if not b:
            return ctx.obj
        return type(ctx.obj)(nm)  # type: ignore

    def rewrite_iterable(self, ctx: Context[IterableT]) -> IterableT:
        nl = []
        b = False
        for i, le in enumerate(ctx.obj):
            ne: ta.Any = self.rewrite(ctx.make_child(le, i))
            nl.append(ne)
            b |= ne is not le
        if not b:
            return ctx.obj
        return type(ctx.obj)(nl)  # type: ignore

    def rewrite_other(self, ctx: Context[T]) -> T:
        return ctx.obj

    def rewrite(self, ctx: Context[T]) -> T:
        if ctx.obj is None:
            return self.rewrite_none(ctx)  # type: ignore

        elif dc.is_dataclass(ctx.obj):
            return self.rewrite_dataclass(ctx)

        elif isinstance(ctx.obj, str):
            return self.rewrite_str(ctx)  # type: ignore

        elif isinstance(ctx.obj, BUILTIN_SCALAR_ITERABLE_TYPES):
            return self.rewrite_builtin_scalar_iterable(ctx)  # type: ignore

        elif isinstance(ctx.obj, collections.abc.Mapping):
            return self.rewrite_mapping(ctx)  # type: ignore

        elif isinstance(ctx.obj, (collections.abc.Sequence, collections.abc.Set)):
            return self.rewrite_iterable(ctx)  # type: ignore

        else:
            return self.rewrite_other(ctx)

    def __call__(self, v: T) -> T:
        return self.rewrite(self.Context(obj=v))



import abc
import functools
import typing as ta

from .abstract import Abstract


##


class FinalTypeError(TypeError):
    def __init__(self, _type: type) -> None:
        super().__init__()

        self._type = _type

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._type})'


class Final(Abstract):
    __slots__ = ()

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        for base in cls.__bases__:
            if base is Abstract:
                raise FinalTypeError(base)
            elif base is Final:
                continue
            elif Final in base.__mro__:
                raise FinalTypeError(base)


##


class SealedError(TypeError):
    def __init__(self, _type) -> None:
        super().__init__()

        self._type = _type

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._type})'


class Sealed:
    __slots__ = ()

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        for base in cls.__bases__:
            if base is not Abstract:
                if Sealed in base.__bases__:
                    if cls.__module__ != base.__module__:
                        raise SealedError(base)

        super().__init_subclass__(**kwargs)


class PackageSealed:
    __slots__ = ()

    __sealed_package__: ta.ClassVar[ta.Sequence[str] | None]

    def __init_subclass__(cls, *, sealed_package: ta.Sequence[str] | str | None = None, **kwargs: ta.Any) -> None:
        if PackageSealed in cls.__bases__:
            if sealed_package is not None:
                if isinstance(sealed_package, str):
                    sealed_package = sealed_package.split('.')
                cls.__sealed_package__ = sealed_package
            else:
                cls.__sealed_package__ = cls.__module__.split('.')[:-1]
        elif sealed_package is not None:
            raise TypeError

        for base in cls.__bases__:
            if (
                    base is Abstract or
                    PackageSealed not in base.__bases__
            ):
                continue
            bsp = base.__dict__['__sealed_package__']
            if cls.__module__.split('.')[:len(bsp)] != bsp:
                raise SealedError(base)

        super().__init_subclass__(**kwargs)


##


class NotInstantiable(Abstract):
    __slots__ = ()

    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError


class NotPicklable:
    __slots__ = ()

    @ta.final
    def __reduce__(self) -> ta.NoReturn:
        raise TypeError

    @ta.final
    def __reduce_ex__(self, protocol) -> ta.NoReturn:
        raise TypeError

    @ta.final
    def __getstate__(self) -> ta.NoReturn:
        raise TypeError

    @ta.final
    def __setstate__(self, state) -> ta.NoReturn:
        raise TypeError


##


class NoBool:
    def __bool__(self) -> bool:
        raise TypeError


class _NoBoolDescriptor:
    def __init__(self, fn, instance=None, owner=None) -> None:
        super().__init__()
        self._fn = fn
        self._instance = instance
        self._owner = owner
        functools.update_wrapper(self, fn)

    def __bool__(self) -> bool:
        raise TypeError

    def __get__(self, instance, owner=None):
        if instance is self._instance and owner is self._owner:
            return self
        return _NoBoolDescriptor(self._fn.__get__(instance, owner), instance, owner)

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)


def no_bool(fn):  # noqa
    return _NoBoolDescriptor(fn)


##


SENSITIVE_ATTR = '__sensitive__'


class Sensitive:
    __sensitive__ = True


class _AnySensitiveMeta(abc.ABCMeta):
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return hasattr(instance, SENSITIVE_ATTR) or super().__instancecheck__(AnySensitive, instance)

    @classmethod
    def __subclasscheck__(cls, subclass: type) -> bool:
        return hasattr(subclass, SENSITIVE_ATTR) or super().__subclasscheck__(AnySensitive, subclass)


class AnySensitive(NotInstantiable, Final, metaclass=_AnySensitiveMeta):
    pass



from .bits import (  # noqa
    get_bit,
    get_bits,
    set_bit,
    set_bits,
)

from .c import (  # noqa
    cdiv,
    cmod,
)

from .fixed import (  # noqa
    CheckedFixedWidthIntError,
    OverflowFixedWidthIntError,
    UnderflowFixedWidthIntError,

    FixedWidthInt,

    SignedInt,
    UnsignedInt,

    CheckedInt,
    ClampedInt,

    AnyInt8,
    AnyInt16,
    AnyInt32,
    AnyInt64,
    AnyInt128,

    CheckedInt8,
    CheckedInt16,
    CheckedInt32,
    CheckedInt64,
    CheckedInt128,

    CheckedUint8,
    CheckedUint16,
    CheckedUint32,
    CheckedUint64,
    CheckedUint128,

    ClampedInt8,
    ClampedInt16,
    ClampedInt32,
    ClampedInt64,
    ClampedInt128,

    ClampedUint8,
    ClampedUint16,
    ClampedUint32,
    ClampedUint64,
    ClampedUint128,

    WrappedInt8,
    WrappedInt16,
    WrappedInt32,
    WrappedInt64,
    WrappedInt128,

    WrappedUint8,
    WrappedUint16,
    WrappedUint32,
    WrappedUint64,
    WrappedUint128,
)

from .floats import (  # noqa
    isclose,
    float_to_bytes,
    bytes_to_float,
)

from .stats import (  # noqa
    get_quantile,

    Stats,
)

from .histogram import (  # noqa
    SamplingHistogram,
)
